import functools
import importlib
import inspect

from lusid_workflows import ApiClient
from lusid_workflows.utilities.api_client_builder import ApiClientBuilder
from lusid_workflows.utilities.proxy_config import ProxyConfig
from lusid_workflows.utilities.api_configuration import ApiConfiguration
from lusid_workflows.utilities.lusid_retry import lusidretry


class ApiClientFactory:
    """
    The ApiClientFactory is responsible for providing the ability to create any of the LUSID APIs using the provided
    credentials. It will use the same ApiClient across all of the APIs.

    """
    def __init__(self, **kwargs):
        """
        Iniitalise an ApiClientFactory by passing the token, api_url and app_name, or by
        passing in the api_secrets_filename

        :param str token: Bearer token used to initialise the API
        :param str api_secrets_filename: Name of secrets file (including full path)
        :param str api_url: LUSID API url
        :param str app_name: Application name (optional)
        :param str certificate_filename: Name of the certificate file (.pem, .cer or .crt)
        :param str proxy_url: The url of the proxy to use including the port e.g. http://myproxy.com:8888
        :param str proxy_username: The username for the proxy to use
        :param str proxy_password: The password for the proxy to use
        :param str correlation_id: Correlation id for all calls made from the returned LUSID API instances
        :param bool tcp_keep_alive: A flag for controlling if the API client uses TCP keep-alive probes
        """

        builder_kwargs = {}

        if "token" in kwargs and str(kwargs["token"]) != "None":
            # If there is a token use it along with the specified proxy details if specified
            config = ApiConfiguration(
                api_url=kwargs.get("api_url", None),
                certificate_filename=kwargs.get("certificate_filename", None),
                proxy_config=ProxyConfig(
                    address=kwargs.get("proxy_url", None),
                    username=kwargs.get("proxy_username", None),
                    password=kwargs.get("proxy_password", None),
                ) if kwargs.get("proxy_url", None) is not None else None,
                app_name=kwargs.get("app_name", None)
            )

            builder_kwargs["api_configuration"] = config
            builder_kwargs["token"] = kwargs["token"]

        # Otherwise use a secrets file if it exists
        builder_kwargs["api_secrets_filename"] = kwargs.get("api_secrets_filename", None)

        # add the correlation id if specified
        builder_kwargs["correlation_id"] = kwargs.get("correlation_id", None)

        # add the id provider response handler if specified
        builder_kwargs["id_provider_response_handler"] = kwargs.get("id_provider_response_handler", None)

        builder_kwargs["tcp_keep_alive"] = kwargs.get("tcp_keep_alive", False)

        # Call the client builder, this will result in using either a token, secrets file or environment variables
        self.api_client = ApiClientBuilder.build(**builder_kwargs)

    def build(self, metaclass):
        """
        :param type metaclass:  type of the LUSID API to create
        :return: Initalised LUSID API for the type passed in
        """

        def get_attribute_impl(source_obj, name):
            """
            Implementation of __getattribute__ that adds a decorator that adds a call_info
            argument to return additional call stats
            """

            attr = super(metaclass, source_obj).__getattribute__(name)

            @functools.wraps(attr)
            @lusidretry
            def wrapper(*args, **kwargs):
                def is_http_info_method(m):
                    return inspect.ismethod(m) and m.__name__.endswith(
                        "_with_http_info"
                    )

                if kwargs.get("call_info") is not None:
                    callback = kwargs.pop("call_info")

                    if not inspect.isfunction(callback):
                        raise ValueError("call_info value must be a lambda")

                    if is_http_info_method(attr):
                        result = attr(*args, **kwargs)
                    else:
                        #   switch to the '_with_http_info' implementation
                        func = getattr(source_obj, f"{name}_with_http_info")
                        result = func(*args, **kwargs)

                    # pass the http info to caller
                    callback(result[2])

                    # return the dto
                    return result[0]

                else:
                    return attr(*args, **kwargs)

            return wrapper if inspect.ismethod(attr) else attr

        def init_impl(dest, src=None):
            if type(dest) == type(src):
                dest.__dict__ = src.__dict__
            elif type(src) == ApiClient:
                dest.api_client = src

        module = importlib.import_module("lusid_workflows.api")
        api_name = metaclass.__name__

        if not api_name.endswith("Api") or not hasattr(module, api_name):
            raise TypeError(f"unknown api: {api_name}")

        # create an instance of the api
        api_impl = getattr(module, api_name)(self.api_client)

        setattr(metaclass, "__getattribute__", get_attribute_impl)
        setattr(metaclass, "__init__", init_impl)

        return api_impl



# coding: utf-8

"""
    FINBOURNE Workflows API

    FINBOURNE Technology  # noqa: E501

    The version of the OpenAPI document: 0.1.21
    Contact: info@finbourne.com
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from lusid_workflows.configuration import Configuration


class WorkflowActivitySummary(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
      required_map (dict): The key is attribute name
                           and the value is whether it is 'required' or 'optional'.
    """
    openapi_types = {
        'activity_name': 'str',
        'activity_type': 'str',
        'activity_execution_status': 'str',
        'started_at': 'datetime',
        'finished_at': 'datetime',
        'execution_info': 'str',
        'activity_context': 'dict(str, object)'
    }

    attribute_map = {
        'activity_name': 'activityName',
        'activity_type': 'activityType',
        'activity_execution_status': 'activityExecutionStatus',
        'started_at': 'startedAt',
        'finished_at': 'finishedAt',
        'execution_info': 'executionInfo',
        'activity_context': 'activityContext'
    }

    required_map = {
        'activity_name': 'optional',
        'activity_type': 'optional',
        'activity_execution_status': 'optional',
        'started_at': 'optional',
        'finished_at': 'optional',
        'execution_info': 'optional',
        'activity_context': 'optional'
    }

    def __init__(self, activity_name=None, activity_type=None, activity_execution_status=None, started_at=None, finished_at=None, execution_info=None, activity_context=None, local_vars_configuration=None):  # noqa: E501
        """WorkflowActivitySummary - a model defined in OpenAPI"
        
        :param activity_name:  Name of the Activity, for example WaitingForFileToReceive, WaitingForIssueToBeAssigned etc
        :type activity_name: str
        :param activity_type:  Type of the Activity, for example Start/StopActivity, ReceiveSignalActivity, SendEmailActivity, ReceiveHttpRequestActivity etc
        :type activity_type: str
        :param activity_execution_status:  Activity's execution status
        :type activity_execution_status: str
        :param started_at:  Timestamp when activity execution began.
        :type started_at: datetime
        :param finished_at:  Timestamp when activity execution finished/faulted/cancelled.
        :type finished_at: datetime
        :param execution_info:  Any info associated with the activity execution. Error message for example.
        :type execution_info: str
        :param activity_context:  Any arbitrary data context associated with the activity
        :type activity_context: dict(str, object)

        """  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._activity_name = None
        self._activity_type = None
        self._activity_execution_status = None
        self._started_at = None
        self._finished_at = None
        self._execution_info = None
        self._activity_context = None
        self.discriminator = None

        self.activity_name = activity_name
        self.activity_type = activity_type
        self.activity_execution_status = activity_execution_status
        self.started_at = started_at
        self.finished_at = finished_at
        self.execution_info = execution_info
        self.activity_context = activity_context

    @property
    def activity_name(self):
        """Gets the activity_name of this WorkflowActivitySummary.  # noqa: E501

        Name of the Activity, for example WaitingForFileToReceive, WaitingForIssueToBeAssigned etc  # noqa: E501

        :return: The activity_name of this WorkflowActivitySummary.  # noqa: E501
        :rtype: str
        """
        return self._activity_name

    @activity_name.setter
    def activity_name(self, activity_name):
        """Sets the activity_name of this WorkflowActivitySummary.

        Name of the Activity, for example WaitingForFileToReceive, WaitingForIssueToBeAssigned etc  # noqa: E501

        :param activity_name: The activity_name of this WorkflowActivitySummary.  # noqa: E501
        :type activity_name: str
        """

        self._activity_name = activity_name

    @property
    def activity_type(self):
        """Gets the activity_type of this WorkflowActivitySummary.  # noqa: E501

        Type of the Activity, for example Start/StopActivity, ReceiveSignalActivity, SendEmailActivity, ReceiveHttpRequestActivity etc  # noqa: E501

        :return: The activity_type of this WorkflowActivitySummary.  # noqa: E501
        :rtype: str
        """
        return self._activity_type

    @activity_type.setter
    def activity_type(self, activity_type):
        """Sets the activity_type of this WorkflowActivitySummary.

        Type of the Activity, for example Start/StopActivity, ReceiveSignalActivity, SendEmailActivity, ReceiveHttpRequestActivity etc  # noqa: E501

        :param activity_type: The activity_type of this WorkflowActivitySummary.  # noqa: E501
        :type activity_type: str
        """

        self._activity_type = activity_type

    @property
    def activity_execution_status(self):
        """Gets the activity_execution_status of this WorkflowActivitySummary.  # noqa: E501

        Activity's execution status  # noqa: E501

        :return: The activity_execution_status of this WorkflowActivitySummary.  # noqa: E501
        :rtype: str
        """
        return self._activity_execution_status

    @activity_execution_status.setter
    def activity_execution_status(self, activity_execution_status):
        """Sets the activity_execution_status of this WorkflowActivitySummary.

        Activity's execution status  # noqa: E501

        :param activity_execution_status: The activity_execution_status of this WorkflowActivitySummary.  # noqa: E501
        :type activity_execution_status: str
        """

        self._activity_execution_status = activity_execution_status

    @property
    def started_at(self):
        """Gets the started_at of this WorkflowActivitySummary.  # noqa: E501

        Timestamp when activity execution began.  # noqa: E501

        :return: The started_at of this WorkflowActivitySummary.  # noqa: E501
        :rtype: datetime
        """
        return self._started_at

    @started_at.setter
    def started_at(self, started_at):
        """Sets the started_at of this WorkflowActivitySummary.

        Timestamp when activity execution began.  # noqa: E501

        :param started_at: The started_at of this WorkflowActivitySummary.  # noqa: E501
        :type started_at: datetime
        """

        self._started_at = started_at

    @property
    def finished_at(self):
        """Gets the finished_at of this WorkflowActivitySummary.  # noqa: E501

        Timestamp when activity execution finished/faulted/cancelled.  # noqa: E501

        :return: The finished_at of this WorkflowActivitySummary.  # noqa: E501
        :rtype: datetime
        """
        return self._finished_at

    @finished_at.setter
    def finished_at(self, finished_at):
        """Sets the finished_at of this WorkflowActivitySummary.

        Timestamp when activity execution finished/faulted/cancelled.  # noqa: E501

        :param finished_at: The finished_at of this WorkflowActivitySummary.  # noqa: E501
        :type finished_at: datetime
        """

        self._finished_at = finished_at

    @property
    def execution_info(self):
        """Gets the execution_info of this WorkflowActivitySummary.  # noqa: E501

        Any info associated with the activity execution. Error message for example.  # noqa: E501

        :return: The execution_info of this WorkflowActivitySummary.  # noqa: E501
        :rtype: str
        """
        return self._execution_info

    @execution_info.setter
    def execution_info(self, execution_info):
        """Sets the execution_info of this WorkflowActivitySummary.

        Any info associated with the activity execution. Error message for example.  # noqa: E501

        :param execution_info: The execution_info of this WorkflowActivitySummary.  # noqa: E501
        :type execution_info: str
        """

        self._execution_info = execution_info

    @property
    def activity_context(self):
        """Gets the activity_context of this WorkflowActivitySummary.  # noqa: E501

        Any arbitrary data context associated with the activity  # noqa: E501

        :return: The activity_context of this WorkflowActivitySummary.  # noqa: E501
        :rtype: dict(str, object)
        """
        return self._activity_context

    @activity_context.setter
    def activity_context(self, activity_context):
        """Sets the activity_context of this WorkflowActivitySummary.

        Any arbitrary data context associated with the activity  # noqa: E501

        :param activity_context: The activity_context of this WorkflowActivitySummary.  # noqa: E501
        :type activity_context: dict(str, object)
        """

        self._activity_context = activity_context

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, WorkflowActivitySummary):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, WorkflowActivitySummary):
            return True

        return self.to_dict() != other.to_dict()



class ProxyConfig:
    """
    This class is used to hold the proxy configuration details
    """

    def __init__(self, address, username=None, password=None):
        """
        :param str address: The address of the proxy including the port e.g. https://myproxy.com:8080
        :param str username: The username for the proxy
        :param str password: The password for the proxy
        """
        self.address = address
        self.__username = username
        self.__password = password

    @property
    def address(self):
        return self.__address

    @address.setter
    def address(self, address):

        if "http://" not in address and "https://" not in address:
            raise ValueError(f"The provided proxy address of {address} does not contain a protocol, please specify in the full format e.g. http://myproxy.com:8080")

        self.__address = address

    @property
    def username(self):
        return self.__username

    @property
    def password(self):
        return self.__password

    def format_proxy_schema(self):
        """
        Takes the proxy details and converts them into a dictionary format to be used with other libraries e.g. requests

        :return: dict: A dictionary with the http and https proxy details including username and password
        """

        proxy_url = self.address

        # Only run if there is a username and password
        if self.username is not None and self.password is not None:
            index = self.address.index("://")

            proxy_url = f"{self.address[0:index + 3]}{self.username}:{self.password}@{self.address[index + 3:]}"

        return {
            "http": proxy_url,
            "https": proxy_url
        }



from lusid_workflows.utilities.api_client_builder import ApiClientBuilder
from lusid_workflows.utilities.api_configuration_loader import ApiConfigurationLoader
from lusid_workflows.utilities.refreshing_token import RefreshingToken
from lusid_workflows.utilities.api_client_factory import ApiClientFactory
from lusid_workflows.utilities.lusid_retry import lusidretry
from lusid_workflows.utilities.proxy_config import ProxyConfig
from lusid_workflows.utilities.api_configuration import ApiConfiguration



# coding: utf-8

"""
    FINBOURNE Workflows API

    FINBOURNE Technology  # noqa: E501

    The version of the OpenAPI document: 0.1.21
    Contact: info@finbourne.com
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from lusid_workflows.configuration import Configuration


class LusidProblemDetails(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
      required_map (dict): The key is attribute name
                           and the value is whether it is 'required' or 'optional'.
    """
    openapi_types = {
        'name': 'str',
        'error_details': 'list[dict(str, str)]',
        'code': 'int',
        'type': 'str',
        'title': 'str',
        'status': 'int',
        'detail': 'str',
        'instance': 'str',
        'extensions': 'dict(str, object)'
    }

    attribute_map = {
        'name': 'name',
        'error_details': 'errorDetails',
        'code': 'code',
        'type': 'type',
        'title': 'title',
        'status': 'status',
        'detail': 'detail',
        'instance': 'instance',
        'extensions': 'extensions'
    }

    required_map = {
        'name': 'required',
        'error_details': 'optional',
        'code': 'required',
        'type': 'optional',
        'title': 'optional',
        'status': 'optional',
        'detail': 'optional',
        'instance': 'optional',
        'extensions': 'optional'
    }

    def __init__(self, name=None, error_details=None, code=None, type=None, title=None, status=None, detail=None, instance=None, extensions=None, local_vars_configuration=None):  # noqa: E501
        """LusidProblemDetails - a model defined in OpenAPI"
        
        :param name:  (required)
        :type name: str
        :param error_details: 
        :type error_details: list[dict(str, str)]
        :param code:  (required)
        :type code: int
        :param type: 
        :type type: str
        :param title: 
        :type title: str
        :param status: 
        :type status: int
        :param detail: 
        :type detail: str
        :param instance: 
        :type instance: str
        :param extensions: 
        :type extensions: dict(str, object)

        """  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._error_details = None
        self._code = None
        self._type = None
        self._title = None
        self._status = None
        self._detail = None
        self._instance = None
        self._extensions = None
        self.discriminator = None

        self.name = name
        self.error_details = error_details
        self.code = code
        self.type = type
        self.title = title
        self.status = status
        self.detail = detail
        self.instance = instance
        self.extensions = extensions

    @property
    def name(self):
        """Gets the name of this LusidProblemDetails.  # noqa: E501


        :return: The name of this LusidProblemDetails.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this LusidProblemDetails.


        :param name: The name of this LusidProblemDetails.  # noqa: E501
        :type name: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def error_details(self):
        """Gets the error_details of this LusidProblemDetails.  # noqa: E501


        :return: The error_details of this LusidProblemDetails.  # noqa: E501
        :rtype: list[dict(str, str)]
        """
        return self._error_details

    @error_details.setter
    def error_details(self, error_details):
        """Sets the error_details of this LusidProblemDetails.


        :param error_details: The error_details of this LusidProblemDetails.  # noqa: E501
        :type error_details: list[dict(str, str)]
        """

        self._error_details = error_details

    @property
    def code(self):
        """Gets the code of this LusidProblemDetails.  # noqa: E501


        :return: The code of this LusidProblemDetails.  # noqa: E501
        :rtype: int
        """
        return self._code

    @code.setter
    def code(self, code):
        """Sets the code of this LusidProblemDetails.


        :param code: The code of this LusidProblemDetails.  # noqa: E501
        :type code: int
        """
        if self.local_vars_configuration.client_side_validation and code is None:  # noqa: E501
            raise ValueError("Invalid value for `code`, must not be `None`")  # noqa: E501

        self._code = code

    @property
    def type(self):
        """Gets the type of this LusidProblemDetails.  # noqa: E501


        :return: The type of this LusidProblemDetails.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this LusidProblemDetails.


        :param type: The type of this LusidProblemDetails.  # noqa: E501
        :type type: str
        """

        self._type = type

    @property
    def title(self):
        """Gets the title of this LusidProblemDetails.  # noqa: E501


        :return: The title of this LusidProblemDetails.  # noqa: E501
        :rtype: str
        """
        return self._title

    @title.setter
    def title(self, title):
        """Sets the title of this LusidProblemDetails.


        :param title: The title of this LusidProblemDetails.  # noqa: E501
        :type title: str
        """

        self._title = title

    @property
    def status(self):
        """Gets the status of this LusidProblemDetails.  # noqa: E501


        :return: The status of this LusidProblemDetails.  # noqa: E501
        :rtype: int
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this LusidProblemDetails.


        :param status: The status of this LusidProblemDetails.  # noqa: E501
        :type status: int
        """

        self._status = status

    @property
    def detail(self):
        """Gets the detail of this LusidProblemDetails.  # noqa: E501


        :return: The detail of this LusidProblemDetails.  # noqa: E501
        :rtype: str
        """
        return self._detail

    @detail.setter
    def detail(self, detail):
        """Sets the detail of this LusidProblemDetails.


        :param detail: The detail of this LusidProblemDetails.  # noqa: E501
        :type detail: str
        """

        self._detail = detail

    @property
    def instance(self):
        """Gets the instance of this LusidProblemDetails.  # noqa: E501


        :return: The instance of this LusidProblemDetails.  # noqa: E501
        :rtype: str
        """
        return self._instance

    @instance.setter
    def instance(self, instance):
        """Sets the instance of this LusidProblemDetails.


        :param instance: The instance of this LusidProblemDetails.  # noqa: E501
        :type instance: str
        """

        self._instance = instance

    @property
    def extensions(self):
        """Gets the extensions of this LusidProblemDetails.  # noqa: E501


        :return: The extensions of this LusidProblemDetails.  # noqa: E501
        :rtype: dict(str, object)
        """
        return self._extensions

    @extensions.setter
    def extensions(self, extensions):
        """Sets the extensions of this LusidProblemDetails.


        :param extensions: The extensions of this LusidProblemDetails.  # noqa: E501
        :type extensions: dict(str, object)
        """

        self._extensions = extensions

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, LusidProblemDetails):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, LusidProblemDetails):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

# flake8: noqa

"""
    FINBOURNE Workflows API

    FINBOURNE Technology  # noqa: E501

    The version of the OpenAPI document: 0.1.21
    Contact: info@finbourne.com
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

__version__ = "0.1.21"

# import apis into sdk package
from lusid_workflows.api.workflows_api import WorkflowsApi

# import ApiClient
from lusid_workflows.api_client import ApiClient
from lusid_workflows.configuration import Configuration
from lusid_workflows.exceptions import OpenApiException
from lusid_workflows.exceptions import ApiTypeError
from lusid_workflows.exceptions import ApiValueError
from lusid_workflows.exceptions import ApiKeyError
from lusid_workflows.exceptions import ApiException
# import models into sdk package
from lusid_workflows.models.execute_workflow_request import ExecuteWorkflowRequest
from lusid_workflows.models.link import Link
from lusid_workflows.models.lusid_problem_details import LusidProblemDetails
from lusid_workflows.models.lusid_validation_problem_details import LusidValidationProblemDetails
from lusid_workflows.models.resource_list_of_workflow import ResourceListOfWorkflow
from lusid_workflows.models.resource_list_of_workflow_instance import ResourceListOfWorkflowInstance
from lusid_workflows.models.transit_workflow_request import TransitWorkflowRequest
from lusid_workflows.models.workflow import Workflow
from lusid_workflows.models.workflow_activity_summary import WorkflowActivitySummary
from lusid_workflows.models.workflow_instance import WorkflowInstance

# import utilities into sdk package
from lusid_workflows.utilities.api_client_builder import ApiClientBuilder
from lusid_workflows.utilities.api_configuration import ApiConfiguration
from lusid_workflows.utilities.api_configuration_loader import ApiConfigurationLoader
from lusid_workflows.utilities.refreshing_token import RefreshingToken

# import tcp utilities
from lusid_workflows.tcp.tcp_keep_alive_probes import TCPKeepAlivePoolManager, TCPKeepAliveProxyManager


import json
import os
from pathlib import Path
import logging

from lusid_workflows.utilities.api_configuration import ApiConfiguration
from lusid_workflows.utilities.proxy_config import ProxyConfig


class ApiConfigurationLoader:
    """
    The ApiConfigurationLoader is responsible for populating the API and Proxy configuration from a secrets file or
    environment variables with preference given to the secrets file.
    """
    @staticmethod
    def load(api_secrets_filename=None):
        """
        :param str api_secrets_filename: The full path to the JSON file containing the API credentials and optional proxy details

        :return: lusid.utilities.ApiConfiguration: The populated ApiConfiguration
        """
        # Get the config keys which contain the mapping between the ApiConfiguration attributes and the variable names
        # in the secrets.json file and environment variables e.g. token_url is tokenUrl (secrets.json) and
        # FBN_TOKEN_URL (env variable)
        with open(Path(__file__).parent.joinpath('config_keys.json')) as json_file:
            config_keys = json.load(json_file)

        # The secrets file is a nested dictionary, set the names of the top level keys
        api_config_key = "api"
        proxy_config_key = "proxy"

        # If there is a secrets file specified and it exists get the details from it
        if api_secrets_filename is not None and os.path.exists(api_secrets_filename) and os.path.isfile(api_secrets_filename):
            with open(api_secrets_filename, "r") as secrets:
                config = json.load(secrets)
        # If there is a secrets file specified and it does not exist log a warning to indicate that the specified file
        # could not be found and create an empty config
        elif api_secrets_filename is not None and (not os.path.exists(api_secrets_filename) or not os.path.isfile(api_secrets_filename)):
            logging.warning(f"Provided secrets file of {api_secrets_filename} can not be found, please ensure you "
                             f"have correctly specified the full path to the file or don't provide a secrets file to use "
                             f"environment variables instead.")
            config = {}
        # If no secrets file is specified just create an empty config
        else:
            config = {}

        # Populate the values for the api configuration preferring the secrets file over the environment variables
        populated_api_config_values = {
            key: config.get(api_config_key, {}).get(value["config"], os.getenv(value["env"], None))
            for key, value in config_keys.items() if "proxy" not in key
        }

        # Populate the values for the proxy preferring the secrets file over the environment variables
        populated_proxy_values = {
            key.replace("proxy_", ""): config.get(proxy_config_key, {}).get(value["config"], os.getenv(value["env"], None))
            for key, value in config_keys.items() if "proxy" in key
        }

        # If the proxy address is missing ensure that no proxy is used in the ApiConfiguration
        if populated_proxy_values.get("address", None) is None:
            populated_api_config_values["proxy_config"] = None
        # Otherwise create a ProxyConfig to use
        else:
            populated_api_config_values["proxy_config"] = ProxyConfig(**populated_proxy_values)
        # Create and return the ApiConfiguration
        return ApiConfiguration(**populated_api_config_values)



# coding: utf-8

"""
    FINBOURNE Workflows API

    FINBOURNE Technology  # noqa: E501

    The version of the OpenAPI document: 0.1.21
    Contact: info@finbourne.com
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from lusid_workflows.configuration import Configuration


class Workflow(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
      required_map (dict): The key is attribute name
                           and the value is whether it is 'required' or 'optional'.
    """
    openapi_types = {
        'description': 'str',
        'code': 'str',
        'latest_version': 'int'
    }

    attribute_map = {
        'description': 'description',
        'code': 'code',
        'latest_version': 'latestVersion'
    }

    required_map = {
        'description': 'optional',
        'code': 'optional',
        'latest_version': 'optional'
    }

    def __init__(self, description=None, code=None, latest_version=None, local_vars_configuration=None):  # noqa: E501
        """Workflow - a model defined in OpenAPI"
        
        :param description:  Workflow Description.
        :type description: str
        :param code:  Friendly workflow code for example IssueWorkflow, PortfolioCreatedWorkflow, TradeBookingWorkflow etc
        :type code: str
        :param latest_version:  Latest workflow definition version. This version increments when workflow definition changes.
        :type latest_version: int

        """  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._description = None
        self._code = None
        self._latest_version = None
        self.discriminator = None

        self.description = description
        self.code = code
        if latest_version is not None:
            self.latest_version = latest_version

    @property
    def description(self):
        """Gets the description of this Workflow.  # noqa: E501

        Workflow Description.  # noqa: E501

        :return: The description of this Workflow.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this Workflow.

        Workflow Description.  # noqa: E501

        :param description: The description of this Workflow.  # noqa: E501
        :type description: str
        """

        self._description = description

    @property
    def code(self):
        """Gets the code of this Workflow.  # noqa: E501

        Friendly workflow code for example IssueWorkflow, PortfolioCreatedWorkflow, TradeBookingWorkflow etc  # noqa: E501

        :return: The code of this Workflow.  # noqa: E501
        :rtype: str
        """
        return self._code

    @code.setter
    def code(self, code):
        """Sets the code of this Workflow.

        Friendly workflow code for example IssueWorkflow, PortfolioCreatedWorkflow, TradeBookingWorkflow etc  # noqa: E501

        :param code: The code of this Workflow.  # noqa: E501
        :type code: str
        """

        self._code = code

    @property
    def latest_version(self):
        """Gets the latest_version of this Workflow.  # noqa: E501

        Latest workflow definition version. This version increments when workflow definition changes.  # noqa: E501

        :return: The latest_version of this Workflow.  # noqa: E501
        :rtype: int
        """
        return self._latest_version

    @latest_version.setter
    def latest_version(self, latest_version):
        """Sets the latest_version of this Workflow.

        Latest workflow definition version. This version increments when workflow definition changes.  # noqa: E501

        :param latest_version: The latest_version of this Workflow.  # noqa: E501
        :type latest_version: int
        """

        self._latest_version = latest_version

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Workflow):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Workflow):
            return True

        return self.to_dict() != other.to_dict()



import requests
import base64
import threading
import time

from datetime import datetime
from datetime import timedelta
from collections import UserString
from urllib.request import pathname2url


class RefreshingToken(UserString):

    def __init__(self, api_configuration, expiry_offset=60, id_provider_response_handler=None):
        """
        Implementation of UserString that will automatically refresh the token value upon expiry

        :param ApiConfiguration api_configuration: The api configuration with all required values
        :param int expiry_offset: number of seconds before token expiry to refresh the token
        :param callable id_provider_response_handler: A handler taking the Requests.Response from the identity provider
        before it is consumed by the RefreshingToken, mutation of the Response is possible with this handler
        """

        self.token_data = {
            "expires": None,
            "access_token": None,
            "refresh_token": None
        }
        self.expiry_offset = expiry_offset
        self.api_configuration = api_configuration
        self.id_provider_response_handler = id_provider_response_handler
        self.refresh_func = self.get_refresh_token
        self.lock = threading.Lock()
        self.retry_count = 0
        self.retry_limit = 5
        self.backoff_base = 2

    def update_token_data(self, id_provider_json):
        """
        Updates the token data from a response from the identity provider

        :param id_provider_json: The JSON to use to update the token data
        """
        self.token_data["access_token"] = id_provider_json["access_token"]
        # Set the expiry just before the actual expiry to ensure no failed requests
        delta = timedelta(seconds=id_provider_json.get("expires_in", 3600) - self.expiry_offset)
        self.token_data["expires"] = datetime.utcnow() + delta
        self.token_data["refresh_token"] = id_provider_json["refresh_token"]

    def get_access_token(self):
        """
        Retrieves an access token from the identity provider using the credentials in the provided configuration

        :return: The retrieved access token
        """

        encoded_password = pathname2url(self.api_configuration.password)
        encoded_client_id = pathname2url(self.api_configuration.client_id)
        encoded_client_secret = pathname2url(self.api_configuration.client_secret)

        # Prepare our authentication request
        token_request_body = f"grant_type=password&username={self.api_configuration.username}" \
            f"&password={encoded_password}&scope=openid client groups offline_access" \
            f"&client_id={encoded_client_id}&client_secret={encoded_client_secret}"

        headers = {"Accept": "application/json", "Content-Type": "application/x-www-form-urlencoded"}

        # extra request args
        kwargs = {"headers": headers}

        if self.api_configuration.proxy_config is not None:
            kwargs["proxies"] = self.api_configuration.proxy_config.format_proxy_schema()

        # use certificate if supplied
        if self.api_configuration.certificate_filename is not None:
            kwargs["verify"] = self.api_configuration.certificate_filename

        # make request to Okta to get an authentication token
        id_provider_response = requests.post(self.api_configuration.token_url, data=token_request_body, **kwargs)

        if self.id_provider_response_handler is not None:
            self.id_provider_response_handler(id_provider_response)

        # Ensure that we have a 200 response code
        if id_provider_response.status_code == 429:
            self._handle_retry(id_provider_response)
            return self.get_access_token()
        elif id_provider_response.status_code != 200:
            raise ValueError(id_provider_response.json())

        self.retry_count = 0

        # convert the json encoded response to be able to extract the token values
        id_provider_json = id_provider_response.json()

        self.update_token_data(id_provider_json)

        return self.token_data["access_token"]

    def get_refresh_token(self):
        """
        Retrieves an access token from the identity provider using the refresh token

        :return: The retrieved access token
        """

        # If any data is missing to use a refresh token e.g. on first try, get an access token using credentials
        if self.token_data["access_token"] is None or self.token_data["expires"] is None or self.token_data["refresh_token"] is None:
            return self.get_access_token()

        # check if the token has expired and refresh if needed
        if self.token_data["expires"] <= datetime.utcnow():

            encoded_client = base64.b64encode(bytes(f"{self.api_configuration.client_id}:{self.api_configuration.client_secret}", 'utf-8'))

            headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                "Authorization": f"Basic {encoded_client.decode('utf-8')}"
            }

            request_body = f"grant_type=refresh_token&scope=openid client groups offline_access&refresh_token={self.token_data['refresh_token']}"

            # request parameters
            kwargs = {"headers": headers}

            if self.api_configuration.proxy_config is not None:
                kwargs["proxies"] = self.api_configuration.proxy_config.format_proxy_schema()

            if self.api_configuration.certificate_filename is not None:
                kwargs["verify"] = self.api_configuration.certificate_filename

            id_provider_response = requests.post(self.api_configuration.token_url, data=request_body, **kwargs)

            if self.id_provider_response_handler is not None:
                self.id_provider_response_handler(id_provider_response)

            # Refresh token may be expired, if so, get new request token
            if id_provider_response.status_code == 400 and 'refresh token is invalid or expired' \
                    in id_provider_response.json()['error_description']:
                return self.get_access_token()
            elif id_provider_response.status_code == 429:
                self._handle_retry(id_provider_response)
                return self.get_refresh_token()
            elif id_provider_response.status_code != 200:
                raise ValueError(id_provider_response.json())

            self.retry_count = 0

            id_provider_json = id_provider_response.json()

            self.update_token_data(id_provider_json)

        return self.token_data["access_token"]

    def _handle_retry(self, id_provider_response):
        """
        Determines how to handle retrying in the event of a failed response. Currently uses the HTTP "Retry-After"
        header to determine how long to wait before retrying. If this header is not present defaults to a simple
        exponential back-off strategy.

        If the identity provider that you are interacting with does not provide a "Retry-After" header but does provide
        other custom headers, you can pass an id_provider_response_handler to the RefreshingToken which constructs
        the "Retry-After" header from the custom headers.

        :param requests.Response id_provider_response: The response from the identity provider
        """
        if self.retry_count >= self.retry_limit:
            raise ValueError(f"Max retry limit of {self.retry_limit} reached with response of {id_provider_response.json()}")

        self.retry_count += 1

        if "Retry-After" in id_provider_response.headers:
            retry_value = id_provider_response.headers.get("Retry-After")

            # Can be a delay in seconds or http date (https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Retry-After)
            try:
                wait_time = int(retry_value)
            except ValueError:
                # http date always in GMT
                wait_time = int(datetime.strptime(retry_value, '%a, %d %b %Y %H:%M:%S GMT').timestamp() - datetime.utcnow().timestamp())
                if wait_time <= 0:  # Won't wait for a negative period
                    return

            time.sleep(wait_time)
            return

        # If no "Retry-After" header implement a simple exponential back-off
        time.sleep(self._calculate_backoff(self.backoff_base, self.retry_count))

    @staticmethod
    def _calculate_backoff(backoff_base, retries):
        """
        Calculates the time to wait before retrying

        :param int retries: The number of retries attempted so far
        :param int backoff_base: The base to use for calculating the backoff

        :return: int: The number of seconds to wait
        """
        return backoff_base ** retries

    def __getattribute__(self, item):

        # return the value of the string
        if item == "data":
            self.lock.acquire()
            # check if the token has expired and go through the refresh token logic if it has
            token = object.__getattribute__(self, "refresh_func")()
            self.lock.release()
            return token

        # get the class attribute to be string class instead of the RefreshingToken class itself, used for UserString
        # base methods such as string concatenation, if this is missing the whole RefreshingToken class gets created
        # again with the concatenated string as the input, with this it creates a string instead
        if item == "__class__":
            return str

        # used to get .self attributes on the RefreshingToken
        return object.__getattribute__(self, item)



# coding: utf-8

"""
    FINBOURNE Workflows API

    FINBOURNE Technology  # noqa: E501

    The version of the OpenAPI document: 0.1.21
    Contact: info@finbourne.com
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import re  # noqa: F401

# python 2 and python 3 compatibility library
import six

from lusid_workflows.api_client import ApiClient
from lusid_workflows.exceptions import (  # noqa: F401
    ApiTypeError,
    ApiValueError
)


class WorkflowsApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

    def execute_workflow(self, workflow_code, execute_workflow_request, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Execute workflow  # noqa: E501

        This will create a new workflow instance for the specified workflow  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.execute_workflow(workflow_code, execute_workflow_request, async_req=True)
        >>> result = thread.get()

        :param workflow_code: Workflow Code (required)
        :type workflow_code: str
        :param execute_workflow_request: Data associated with execute request (required)
        :type execute_workflow_request: ExecuteWorkflowRequest
        :param correlation_id: Optional. This id can be used to correlate a workflow instance with any entity (e.g. a custom entity or other workflow instance) and also to retrieve and transit workflow instances
        :type correlation_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkflowInstance
        """
        kwargs['_return_http_data_only'] = True
        return self.execute_workflow_with_http_info(workflow_code, execute_workflow_request, **kwargs)  # noqa: E501

    def execute_workflow_with_http_info(self, workflow_code, execute_workflow_request, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Execute workflow  # noqa: E501

        This will create a new workflow instance for the specified workflow  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.execute_workflow_with_http_info(workflow_code, execute_workflow_request, async_req=True)
        >>> result = thread.get()

        :param workflow_code: Workflow Code (required)
        :type workflow_code: str
        :param execute_workflow_request: Data associated with execute request (required)
        :type execute_workflow_request: ExecuteWorkflowRequest
        :param correlation_id: Optional. This id can be used to correlate a workflow instance with any entity (e.g. a custom entity or other workflow instance) and also to retrieve and transit workflow instances
        :type correlation_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(WorkflowInstance, status_code(int), headers(HTTPHeaderDict))
        """

        local_var_params = locals()

        all_params = [
            'workflow_code',
            'execute_workflow_request',
            'correlation_id'
        ]
        all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth'
            ]
        )

        for key, val in six.iteritems(local_var_params['kwargs']):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method execute_workflow" % key
                )
            local_var_params[key] = val
        del local_var_params['kwargs']
        # verify the required parameter 'workflow_code' is set
        if self.api_client.client_side_validation and ('workflow_code' not in local_var_params or  # noqa: E501
                                                        local_var_params['workflow_code'] is None):  # noqa: E501
            raise ApiValueError("Missing the required parameter `workflow_code` when calling `execute_workflow`")  # noqa: E501
        # verify the required parameter 'execute_workflow_request' is set
        if self.api_client.client_side_validation and ('execute_workflow_request' not in local_var_params or  # noqa: E501
                                                        local_var_params['execute_workflow_request'] is None):  # noqa: E501
            raise ApiValueError("Missing the required parameter `execute_workflow_request` when calling `execute_workflow`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'workflow_code' in local_var_params:
            path_params['workflowCode'] = local_var_params['workflow_code']  # noqa: E501

        query_params = []
        if 'correlation_id' in local_var_params and local_var_params['correlation_id'] is not None:  # noqa: E501
            query_params.append(('correlationId', local_var_params['correlation_id']))  # noqa: E501

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'execute_workflow_request' in local_var_params:
            body_params = local_var_params['execute_workflow_request']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        header_params['Accept-Encoding'] = "gzip, deflate, br"

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json-patch+json', 'application/json', 'text/json', 'application/*+json'])  # noqa: E501

        # set the LUSID header
        header_params['X-LUSID-SDK-Language'] = 'Python'
        header_params['X-LUSID-SDK-Version'] = '0.1.21'

        # Authentication setting
        auth_settings = ['oauth2']  # noqa: E501
        
        response_types_map = {
            200: "WorkflowInstance",
            400: "LusidValidationProblemDetails",
            404: "str",
        }

        return self.api_client.call_api(
            '/api/v1/workflows/{workflowCode}', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_types_map=response_types_map,
            auth_settings=auth_settings,
            async_req=local_var_params.get('async_req'),
            _return_http_data_only=local_var_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=local_var_params.get('_preload_content', True),
            _request_timeout=local_var_params.get('_request_timeout'),
            collection_formats=collection_formats,
            _request_auth=local_var_params.get('_request_auth'))

    def get_workflow_instance(self, workflow_instance_id, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Get workflow instance  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_workflow_instance(workflow_instance_id, async_req=True)
        >>> result = thread.get()

        :param workflow_instance_id: Unique workflow instance identifier (required)
        :type workflow_instance_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkflowInstance
        """
        kwargs['_return_http_data_only'] = True
        return self.get_workflow_instance_with_http_info(workflow_instance_id, **kwargs)  # noqa: E501

    def get_workflow_instance_with_http_info(self, workflow_instance_id, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Get workflow instance  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_workflow_instance_with_http_info(workflow_instance_id, async_req=True)
        >>> result = thread.get()

        :param workflow_instance_id: Unique workflow instance identifier (required)
        :type workflow_instance_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(WorkflowInstance, status_code(int), headers(HTTPHeaderDict))
        """

        local_var_params = locals()

        all_params = [
            'workflow_instance_id'
        ]
        all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth'
            ]
        )

        for key, val in six.iteritems(local_var_params['kwargs']):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_workflow_instance" % key
                )
            local_var_params[key] = val
        del local_var_params['kwargs']
        # verify the required parameter 'workflow_instance_id' is set
        if self.api_client.client_side_validation and ('workflow_instance_id' not in local_var_params or  # noqa: E501
                                                        local_var_params['workflow_instance_id'] is None):  # noqa: E501
            raise ApiValueError("Missing the required parameter `workflow_instance_id` when calling `get_workflow_instance`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'workflow_instance_id' in local_var_params:
            path_params['workflowInstanceId'] = local_var_params['workflow_instance_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        header_params['Accept-Encoding'] = "gzip, deflate, br"

        # Authentication setting
        auth_settings = ['oauth2']  # noqa: E501
        
        response_types_map = {
            200: "WorkflowInstance",
            400: "LusidValidationProblemDetails",
            404: "str",
        }

        return self.api_client.call_api(
            '/api/v1/workflows/instances/{workflowInstanceId}', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_types_map=response_types_map,
            auth_settings=auth_settings,
            async_req=local_var_params.get('async_req'),
            _return_http_data_only=local_var_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=local_var_params.get('_preload_content', True),
            _request_timeout=local_var_params.get('_request_timeout'),
            collection_formats=collection_formats,
            _request_auth=local_var_params.get('_request_auth'))

    def get_workflow_instance_by_correlation_id(self, correlation_id, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Get workflow instance  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_workflow_instance_by_correlation_id(correlation_id, async_req=True)
        >>> result = thread.get()

        :param correlation_id: CorrelationId associated with the workflow instance (required)
        :type correlation_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkflowInstance
        """
        kwargs['_return_http_data_only'] = True
        return self.get_workflow_instance_by_correlation_id_with_http_info(correlation_id, **kwargs)  # noqa: E501

    def get_workflow_instance_by_correlation_id_with_http_info(self, correlation_id, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Get workflow instance  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_workflow_instance_by_correlation_id_with_http_info(correlation_id, async_req=True)
        >>> result = thread.get()

        :param correlation_id: CorrelationId associated with the workflow instance (required)
        :type correlation_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(WorkflowInstance, status_code(int), headers(HTTPHeaderDict))
        """

        local_var_params = locals()

        all_params = [
            'correlation_id'
        ]
        all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth'
            ]
        )

        for key, val in six.iteritems(local_var_params['kwargs']):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_workflow_instance_by_correlation_id" % key
                )
            local_var_params[key] = val
        del local_var_params['kwargs']
        # verify the required parameter 'correlation_id' is set
        if self.api_client.client_side_validation and ('correlation_id' not in local_var_params or  # noqa: E501
                                                        local_var_params['correlation_id'] is None):  # noqa: E501
            raise ApiValueError("Missing the required parameter `correlation_id` when calling `get_workflow_instance_by_correlation_id`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'correlation_id' in local_var_params:
            path_params['correlationId'] = local_var_params['correlation_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        header_params['Accept-Encoding'] = "gzip, deflate, br"

        # Authentication setting
        auth_settings = ['oauth2']  # noqa: E501
        
        response_types_map = {
            200: "WorkflowInstance",
            400: "LusidValidationProblemDetails",
            404: "str",
        }

        return self.api_client.call_api(
            '/api/v1/workflows/instance/{correlationId}', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_types_map=response_types_map,
            auth_settings=auth_settings,
            async_req=local_var_params.get('async_req'),
            _return_http_data_only=local_var_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=local_var_params.get('_preload_content', True),
            _request_timeout=local_var_params.get('_request_timeout'),
            collection_formats=collection_formats,
            _request_auth=local_var_params.get('_request_auth'))

    def list_workflow_instances(self, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Get workflow instances  # noqa: E501

        If no parameters are passed then all instances will be returned  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_workflow_instances(async_req=True)
        >>> result = thread.get()

        :param workflow_code: Optional. Workflow code
        :type workflow_code: str
        :param workflow_execution_status: Optional. Workflow execution status
        :type workflow_execution_status: str
        :param workflow_instance_state: Optional. Custom state associated with workflow instance
        :type workflow_instance_state: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ResourceListOfWorkflowInstance
        """
        kwargs['_return_http_data_only'] = True
        return self.list_workflow_instances_with_http_info(**kwargs)  # noqa: E501

    def list_workflow_instances_with_http_info(self, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Get workflow instances  # noqa: E501

        If no parameters are passed then all instances will be returned  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_workflow_instances_with_http_info(async_req=True)
        >>> result = thread.get()

        :param workflow_code: Optional. Workflow code
        :type workflow_code: str
        :param workflow_execution_status: Optional. Workflow execution status
        :type workflow_execution_status: str
        :param workflow_instance_state: Optional. Custom state associated with workflow instance
        :type workflow_instance_state: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(ResourceListOfWorkflowInstance, status_code(int), headers(HTTPHeaderDict))
        """

        local_var_params = locals()

        all_params = [
            'workflow_code',
            'workflow_execution_status',
            'workflow_instance_state'
        ]
        all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth'
            ]
        )

        for key, val in six.iteritems(local_var_params['kwargs']):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method list_workflow_instances" % key
                )
            local_var_params[key] = val
        del local_var_params['kwargs']

        collection_formats = {}

        path_params = {}

        query_params = []
        if 'workflow_code' in local_var_params and local_var_params['workflow_code'] is not None:  # noqa: E501
            query_params.append(('workflowCode', local_var_params['workflow_code']))  # noqa: E501
        if 'workflow_execution_status' in local_var_params and local_var_params['workflow_execution_status'] is not None:  # noqa: E501
            query_params.append(('workflowExecutionStatus', local_var_params['workflow_execution_status']))  # noqa: E501
        if 'workflow_instance_state' in local_var_params and local_var_params['workflow_instance_state'] is not None:  # noqa: E501
            query_params.append(('workflowInstanceState', local_var_params['workflow_instance_state']))  # noqa: E501

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        header_params['Accept-Encoding'] = "gzip, deflate, br"

        # Authentication setting
        auth_settings = ['oauth2']  # noqa: E501
        
        response_types_map = {
            200: "ResourceListOfWorkflowInstance",
            400: "LusidValidationProblemDetails",
        }

        return self.api_client.call_api(
            '/api/v1/workflows/instances', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_types_map=response_types_map,
            auth_settings=auth_settings,
            async_req=local_var_params.get('async_req'),
            _return_http_data_only=local_var_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=local_var_params.get('_preload_content', True),
            _request_timeout=local_var_params.get('_request_timeout'),
            collection_formats=collection_formats,
            _request_auth=local_var_params.get('_request_auth'))

    def list_workflows(self, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Get all available workflows  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_workflows(async_req=True)
        >>> result = thread.get()

        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ResourceListOfWorkflow
        """
        kwargs['_return_http_data_only'] = True
        return self.list_workflows_with_http_info(**kwargs)  # noqa: E501

    def list_workflows_with_http_info(self, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Get all available workflows  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_workflows_with_http_info(async_req=True)
        >>> result = thread.get()

        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(ResourceListOfWorkflow, status_code(int), headers(HTTPHeaderDict))
        """

        local_var_params = locals()

        all_params = [
        ]
        all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth'
            ]
        )

        for key, val in six.iteritems(local_var_params['kwargs']):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method list_workflows" % key
                )
            local_var_params[key] = val
        del local_var_params['kwargs']

        collection_formats = {}

        path_params = {}

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        header_params['Accept-Encoding'] = "gzip, deflate, br"

        # Authentication setting
        auth_settings = ['oauth2']  # noqa: E501
        
        response_types_map = {
            200: "ResourceListOfWorkflow",
        }

        return self.api_client.call_api(
            '/api/v1/workflows', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_types_map=response_types_map,
            auth_settings=auth_settings,
            async_req=local_var_params.get('async_req'),
            _return_http_data_only=local_var_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=local_var_params.get('_preload_content', True),
            _request_timeout=local_var_params.get('_request_timeout'),
            collection_formats=collection_formats,
            _request_auth=local_var_params.get('_request_auth'))

    def transit_workflow(self, workflow_instance_id, transit_workflow_request, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Transit workflow  # noqa: E501

        This would transit the workflow instance to proceed to next step using WorkflowInstanceId with the attached data/context and action  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.transit_workflow(workflow_instance_id, transit_workflow_request, async_req=True)
        >>> result = thread.get()

        :param workflow_instance_id: Unique workflow instance identifier (required)
        :type workflow_instance_id: str
        :param transit_workflow_request: Data associated with the transit request (required)
        :type transit_workflow_request: TransitWorkflowRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkflowInstance
        """
        kwargs['_return_http_data_only'] = True
        return self.transit_workflow_with_http_info(workflow_instance_id, transit_workflow_request, **kwargs)  # noqa: E501

    def transit_workflow_with_http_info(self, workflow_instance_id, transit_workflow_request, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Transit workflow  # noqa: E501

        This would transit the workflow instance to proceed to next step using WorkflowInstanceId with the attached data/context and action  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.transit_workflow_with_http_info(workflow_instance_id, transit_workflow_request, async_req=True)
        >>> result = thread.get()

        :param workflow_instance_id: Unique workflow instance identifier (required)
        :type workflow_instance_id: str
        :param transit_workflow_request: Data associated with the transit request (required)
        :type transit_workflow_request: TransitWorkflowRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(WorkflowInstance, status_code(int), headers(HTTPHeaderDict))
        """

        local_var_params = locals()

        all_params = [
            'workflow_instance_id',
            'transit_workflow_request'
        ]
        all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth'
            ]
        )

        for key, val in six.iteritems(local_var_params['kwargs']):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method transit_workflow" % key
                )
            local_var_params[key] = val
        del local_var_params['kwargs']
        # verify the required parameter 'workflow_instance_id' is set
        if self.api_client.client_side_validation and ('workflow_instance_id' not in local_var_params or  # noqa: E501
                                                        local_var_params['workflow_instance_id'] is None):  # noqa: E501
            raise ApiValueError("Missing the required parameter `workflow_instance_id` when calling `transit_workflow`")  # noqa: E501
        # verify the required parameter 'transit_workflow_request' is set
        if self.api_client.client_side_validation and ('transit_workflow_request' not in local_var_params or  # noqa: E501
                                                        local_var_params['transit_workflow_request'] is None):  # noqa: E501
            raise ApiValueError("Missing the required parameter `transit_workflow_request` when calling `transit_workflow`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'workflow_instance_id' in local_var_params:
            path_params['workflowInstanceId'] = local_var_params['workflow_instance_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'transit_workflow_request' in local_var_params:
            body_params = local_var_params['transit_workflow_request']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        header_params['Accept-Encoding'] = "gzip, deflate, br"

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json-patch+json', 'application/json', 'text/json', 'application/*+json'])  # noqa: E501

        # set the LUSID header
        header_params['X-LUSID-SDK-Language'] = 'Python'
        header_params['X-LUSID-SDK-Version'] = '0.1.21'

        # Authentication setting
        auth_settings = ['oauth2']  # noqa: E501
        
        response_types_map = {
            200: "WorkflowInstance",
            400: "LusidValidationProblemDetails",
            404: "str",
        }

        return self.api_client.call_api(
            '/api/v1/workflows/instances/{workflowInstanceId}', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_types_map=response_types_map,
            auth_settings=auth_settings,
            async_req=local_var_params.get('async_req'),
            _return_http_data_only=local_var_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=local_var_params.get('_preload_content', True),
            _request_timeout=local_var_params.get('_request_timeout'),
            collection_formats=collection_formats,
            _request_auth=local_var_params.get('_request_auth'))

    def transit_workflow_by_correlation_id(self, correlation_id, transit_workflow_request, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Transit workflow  # noqa: E501

        This would transit the workflow instance to proceed to next step using CorrelationId with the attached data/context and action  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.transit_workflow_by_correlation_id(correlation_id, transit_workflow_request, async_req=True)
        >>> result = thread.get()

        :param correlation_id: CorrelationId associated with the workflow instance (required)
        :type correlation_id: str
        :param transit_workflow_request: Data associated with the transit request (required)
        :type transit_workflow_request: TransitWorkflowRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkflowInstance
        """
        kwargs['_return_http_data_only'] = True
        return self.transit_workflow_by_correlation_id_with_http_info(correlation_id, transit_workflow_request, **kwargs)  # noqa: E501

    def transit_workflow_by_correlation_id_with_http_info(self, correlation_id, transit_workflow_request, **kwargs):  # noqa: E501
        """[EXPERIMENTAL] Transit workflow  # noqa: E501

        This would transit the workflow instance to proceed to next step using CorrelationId with the attached data/context and action  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.transit_workflow_by_correlation_id_with_http_info(correlation_id, transit_workflow_request, async_req=True)
        >>> result = thread.get()

        :param correlation_id: CorrelationId associated with the workflow instance (required)
        :type correlation_id: str
        :param transit_workflow_request: Data associated with the transit request (required)
        :type transit_workflow_request: TransitWorkflowRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(WorkflowInstance, status_code(int), headers(HTTPHeaderDict))
        """

        local_var_params = locals()

        all_params = [
            'correlation_id',
            'transit_workflow_request'
        ]
        all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth'
            ]
        )

        for key, val in six.iteritems(local_var_params['kwargs']):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method transit_workflow_by_correlation_id" % key
                )
            local_var_params[key] = val
        del local_var_params['kwargs']
        # verify the required parameter 'correlation_id' is set
        if self.api_client.client_side_validation and ('correlation_id' not in local_var_params or  # noqa: E501
                                                        local_var_params['correlation_id'] is None):  # noqa: E501
            raise ApiValueError("Missing the required parameter `correlation_id` when calling `transit_workflow_by_correlation_id`")  # noqa: E501
        # verify the required parameter 'transit_workflow_request' is set
        if self.api_client.client_side_validation and ('transit_workflow_request' not in local_var_params or  # noqa: E501
                                                        local_var_params['transit_workflow_request'] is None):  # noqa: E501
            raise ApiValueError("Missing the required parameter `transit_workflow_request` when calling `transit_workflow_by_correlation_id`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'correlation_id' in local_var_params:
            path_params['correlationId'] = local_var_params['correlation_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'transit_workflow_request' in local_var_params:
            body_params = local_var_params['transit_workflow_request']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        header_params['Accept-Encoding'] = "gzip, deflate, br"

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json-patch+json', 'application/json', 'text/json', 'application/*+json'])  # noqa: E501

        # set the LUSID header
        header_params['X-LUSID-SDK-Language'] = 'Python'
        header_params['X-LUSID-SDK-Version'] = '0.1.21'

        # Authentication setting
        auth_settings = ['oauth2']  # noqa: E501
        
        response_types_map = {
            200: "WorkflowInstance",
            400: "LusidValidationProblemDetails",
            404: "str",
        }

        return self.api_client.call_api(
            '/api/v1/workflows/instance/{correlationId}', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_types_map=response_types_map,
            auth_settings=auth_settings,
            async_req=local_var_params.get('async_req'),
            _return_http_data_only=local_var_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=local_var_params.get('_preload_content', True),
            _request_timeout=local_var_params.get('_request_timeout'),
            collection_formats=collection_formats,
            _request_auth=local_var_params.get('_request_auth'))



from urllib3 import HTTPSConnectionPool, HTTPConnectionPool, PoolManager, ProxyManager
import socket
import sys

# The content to send on Mac OS in the TCP Keep Alive probe
TCP_KEEPALIVE = 0x10
# The maximum time to keep the connection idle before sending probes
TCP_KEEP_IDLE = 60
# The interval between probes
TCP_KEEPALIVE_INTERVAL = 60
# The maximum number of failed probes before terminating the connection
TCP_KEEP_CNT = 3


class TCPKeepAliveValidationMethods:
    """
    This class contains a single method whose sole purpose is to set up TCP Keep Alive probes on the socket for a
    connection. This is necessary for long running requests which will be silently terminated by the AWS Network Load
    Balancer which kills a connection if it is idle for more then 350 seconds.
    """
    @staticmethod
    def adjust_connection_socket(conn, protocol: str = "https"):
        """
        Adjusts the socket settings so that the client sends a TCP keep alive probe over the connection. This is only
        applied where possible, if the ability to set the socket options is not available, for example using Anaconda,
        then the settings will be left as is.

        :param conn: The connection to update the socket settings for
        :param str protocol: The protocol of the connection

        :return: None
        """

        if protocol == "http":
            # It isn't clear how to set this up over HTTP, it seems to differ from HTTPs
            return

        # TCP Keep Alive Probes for different platforms
        platform = sys.platform
        # TCP Keep Alive Probes for Linux
        if (platform == 'linux' and hasattr(conn.sock, "setsockopt") and hasattr(socket, "SO_KEEPALIVE") and
                hasattr(socket, "TCP_KEEPIDLE") and hasattr(socket, "TCP_KEEPINTVL") and hasattr(socket, "TCP_KEEPCNT")):
            conn.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            conn.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, TCP_KEEP_IDLE)
            conn.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, TCP_KEEPALIVE_INTERVAL)
            conn.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, TCP_KEEP_CNT)

        # TCP Keep Alive Probes for Windows OS
        elif platform == 'win32' and hasattr(socket, "SIO_KEEPALIVE_VALS") and getattr(conn.sock, "ioctl", None) is not None:
            conn.sock.ioctl(socket.SIO_KEEPALIVE_VALS, (1, TCP_KEEP_IDLE * 1000, TCP_KEEPALIVE_INTERVAL * 1000))

        # TCP Keep Alive Probes for Mac OS
        elif platform == 'darwin' and getattr(conn.sock, "setsockopt", None) is not None:
            conn.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            conn.sock.setsockopt(socket.IPPROTO_TCP, TCP_KEEPALIVE, TCP_KEEPALIVE_INTERVAL)


class TCPKeepAliveHTTPSConnectionPool(HTTPSConnectionPool):
    """
    This class overrides the _validate_conn method in the HTTPSConnectionPool class. This is the entry point to use
    for modifying the socket as it is called after the socket is created and before the request is made.
    """
    def _validate_conn(self, conn):
        """
        Called right before a request is made, after the socket is created.
        """
        # Call the method on the base class
        super()._validate_conn(conn)

        # Set up TCP Keep Alive probes, this is the only line added to this function
        TCPKeepAliveValidationMethods.adjust_connection_socket(conn, "https")


class TCPKeepAliveHTTPConnectionPool(HTTPConnectionPool):
    """
    This class overrides the _validate_conn method in the HTTPSConnectionPool class. This is the entry point to use
    for modifying the socket as it is called after the socket is created and before the request is made.

    In the base class this method is passed completely.
    """

    def _validate_conn(self, conn):
        """
        Called right before a request is made, after the socket is created.
        """
        # Call the method on the base class
        super()._validate_conn(conn)

        # Set up TCP Keep Alive probes, this is the only line added to this function
        TCPKeepAliveValidationMethods.adjust_connection_socket(conn, "http")


class TCPKeepAlivePoolManager(PoolManager):
    """
    This Pool Manager has only had the pool_classes_by_scheme variable changed. This now points at the TCPKeepAlive
    connection pools rather than the default connection pools.
    """
    def __init__(self, num_pools=10, headers=None, **connection_pool_kw):
        super().__init__(num_pools=num_pools, headers=headers, **connection_pool_kw)
        self.pool_classes_by_scheme = {"http": TCPKeepAliveHTTPConnectionPool, "https": TCPKeepAliveHTTPSConnectionPool}


class TCPKeepAliveProxyManager(ProxyManager):
    """
    This Proxy Manager has only had the pool_classes_by_scheme variable changed. This now points at the TCPKeepAlive
    connection pools rather than the default connection pools.
    """
    def __init__(self, proxy_url, num_pools=10, headers=None, proxy_headers=None, **connection_pool_kw):
        super().__init__(proxy_url=proxy_url, num_pools=num_pools, headers=headers, proxy_headers=proxy_headers, **connection_pool_kw)
        self.pool_classes_by_scheme = {"http": TCPKeepAliveHTTPConnectionPool, "https": TCPKeepAliveHTTPSConnectionPool}


# coding: utf-8

"""
    FINBOURNE Workflows API

    FINBOURNE Technology  # noqa: E501

    The version of the OpenAPI document: 0.1.21
    Contact: info@finbourne.com
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from lusid_workflows.configuration import Configuration


class Link(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
      required_map (dict): The key is attribute name
                           and the value is whether it is 'required' or 'optional'.
    """
    openapi_types = {
        'relation': 'str',
        'href': 'str',
        'description': 'str',
        'method': 'str'
    }

    attribute_map = {
        'relation': 'relation',
        'href': 'href',
        'description': 'description',
        'method': 'method'
    }

    required_map = {
        'relation': 'required',
        'href': 'required',
        'description': 'optional',
        'method': 'required'
    }

    def __init__(self, relation=None, href=None, description=None, method=None, local_vars_configuration=None):  # noqa: E501
        """Link - a model defined in OpenAPI"
        
        :param relation:  (required)
        :type relation: str
        :param href:  (required)
        :type href: str
        :param description: 
        :type description: str
        :param method:  (required)
        :type method: str

        """  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._relation = None
        self._href = None
        self._description = None
        self._method = None
        self.discriminator = None

        self.relation = relation
        self.href = href
        self.description = description
        self.method = method

    @property
    def relation(self):
        """Gets the relation of this Link.  # noqa: E501


        :return: The relation of this Link.  # noqa: E501
        :rtype: str
        """
        return self._relation

    @relation.setter
    def relation(self, relation):
        """Sets the relation of this Link.


        :param relation: The relation of this Link.  # noqa: E501
        :type relation: str
        """
        if self.local_vars_configuration.client_side_validation and relation is None:  # noqa: E501
            raise ValueError("Invalid value for `relation`, must not be `None`")  # noqa: E501

        self._relation = relation

    @property
    def href(self):
        """Gets the href of this Link.  # noqa: E501


        :return: The href of this Link.  # noqa: E501
        :rtype: str
        """
        return self._href

    @href.setter
    def href(self, href):
        """Sets the href of this Link.


        :param href: The href of this Link.  # noqa: E501
        :type href: str
        """
        if self.local_vars_configuration.client_side_validation and href is None:  # noqa: E501
            raise ValueError("Invalid value for `href`, must not be `None`")  # noqa: E501

        self._href = href

    @property
    def description(self):
        """Gets the description of this Link.  # noqa: E501


        :return: The description of this Link.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this Link.


        :param description: The description of this Link.  # noqa: E501
        :type description: str
        """

        self._description = description

    @property
    def method(self):
        """Gets the method of this Link.  # noqa: E501


        :return: The method of this Link.  # noqa: E501
        :rtype: str
        """
        return self._method

    @method.setter
    def method(self, method):
        """Sets the method of this Link.


        :param method: The method of this Link.  # noqa: E501
        :type method: str
        """
        if self.local_vars_configuration.client_side_validation and method is None:  # noqa: E501
            raise ValueError("Invalid value for `method`, must not be `None`")  # noqa: E501

        self._method = method

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Link):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Link):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

# flake8: noqa
"""
    FINBOURNE Workflows API

    FINBOURNE Technology  # noqa: E501

    The version of the OpenAPI document: 0.1.21
    Contact: info@finbourne.com
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

# import models into model package
from lusid_workflows.models.execute_workflow_request import ExecuteWorkflowRequest
from lusid_workflows.models.link import Link
from lusid_workflows.models.lusid_problem_details import LusidProblemDetails
from lusid_workflows.models.lusid_validation_problem_details import LusidValidationProblemDetails
from lusid_workflows.models.resource_list_of_workflow import ResourceListOfWorkflow
from lusid_workflows.models.resource_list_of_workflow_instance import ResourceListOfWorkflowInstance
from lusid_workflows.models.transit_workflow_request import TransitWorkflowRequest
from lusid_workflows.models.workflow import Workflow
from lusid_workflows.models.workflow_activity_summary import WorkflowActivitySummary
from lusid_workflows.models.workflow_instance import WorkflowInstance



# coding: utf-8

"""
    FINBOURNE Workflows API

    FINBOURNE Technology  # noqa: E501

    The version of the OpenAPI document: 0.1.21
    Contact: info@finbourne.com
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from lusid_workflows.configuration import Configuration


class ResourceListOfWorkflowInstance(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
      required_map (dict): The key is attribute name
                           and the value is whether it is 'required' or 'optional'.
    """
    openapi_types = {
        'values': 'list[WorkflowInstance]',
        'href': 'str',
        'links': 'list[Link]',
        'next_page': 'str',
        'previous_page': 'str'
    }

    attribute_map = {
        'values': 'values',
        'href': 'href',
        'links': 'links',
        'next_page': 'nextPage',
        'previous_page': 'previousPage'
    }

    required_map = {
        'values': 'required',
        'href': 'optional',
        'links': 'optional',
        'next_page': 'optional',
        'previous_page': 'optional'
    }

    def __init__(self, values=None, href=None, links=None, next_page=None, previous_page=None, local_vars_configuration=None):  # noqa: E501
        """ResourceListOfWorkflowInstance - a model defined in OpenAPI"
        
        :param values:  (required)
        :type values: list[lusid_workflows.WorkflowInstance]
        :param href: 
        :type href: str
        :param links: 
        :type links: list[lusid_workflows.Link]
        :param next_page: 
        :type next_page: str
        :param previous_page: 
        :type previous_page: str

        """  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._values = None
        self._href = None
        self._links = None
        self._next_page = None
        self._previous_page = None
        self.discriminator = None

        self.values = values
        self.href = href
        self.links = links
        self.next_page = next_page
        self.previous_page = previous_page

    @property
    def values(self):
        """Gets the values of this ResourceListOfWorkflowInstance.  # noqa: E501


        :return: The values of this ResourceListOfWorkflowInstance.  # noqa: E501
        :rtype: list[lusid_workflows.WorkflowInstance]
        """
        return self._values

    @values.setter
    def values(self, values):
        """Sets the values of this ResourceListOfWorkflowInstance.


        :param values: The values of this ResourceListOfWorkflowInstance.  # noqa: E501
        :type values: list[lusid_workflows.WorkflowInstance]
        """
        if self.local_vars_configuration.client_side_validation and values is None:  # noqa: E501
            raise ValueError("Invalid value for `values`, must not be `None`")  # noqa: E501

        self._values = values

    @property
    def href(self):
        """Gets the href of this ResourceListOfWorkflowInstance.  # noqa: E501


        :return: The href of this ResourceListOfWorkflowInstance.  # noqa: E501
        :rtype: str
        """
        return self._href

    @href.setter
    def href(self, href):
        """Sets the href of this ResourceListOfWorkflowInstance.


        :param href: The href of this ResourceListOfWorkflowInstance.  # noqa: E501
        :type href: str
        """

        self._href = href

    @property
    def links(self):
        """Gets the links of this ResourceListOfWorkflowInstance.  # noqa: E501


        :return: The links of this ResourceListOfWorkflowInstance.  # noqa: E501
        :rtype: list[lusid_workflows.Link]
        """
        return self._links

    @links.setter
    def links(self, links):
        """Sets the links of this ResourceListOfWorkflowInstance.


        :param links: The links of this ResourceListOfWorkflowInstance.  # noqa: E501
        :type links: list[lusid_workflows.Link]
        """

        self._links = links

    @property
    def next_page(self):
        """Gets the next_page of this ResourceListOfWorkflowInstance.  # noqa: E501


        :return: The next_page of this ResourceListOfWorkflowInstance.  # noqa: E501
        :rtype: str
        """
        return self._next_page

    @next_page.setter
    def next_page(self, next_page):
        """Sets the next_page of this ResourceListOfWorkflowInstance.


        :param next_page: The next_page of this ResourceListOfWorkflowInstance.  # noqa: E501
        :type next_page: str
        """

        self._next_page = next_page

    @property
    def previous_page(self):
        """Gets the previous_page of this ResourceListOfWorkflowInstance.  # noqa: E501


        :return: The previous_page of this ResourceListOfWorkflowInstance.  # noqa: E501
        :rtype: str
        """
        return self._previous_page

    @previous_page.setter
    def previous_page(self, previous_page):
        """Sets the previous_page of this ResourceListOfWorkflowInstance.


        :param previous_page: The previous_page of this ResourceListOfWorkflowInstance.  # noqa: E501
        :type previous_page: str
        """

        self._previous_page = previous_page

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResourceListOfWorkflowInstance):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResourceListOfWorkflowInstance):
            return True

        return self.to_dict() != other.to_dict()



import functools
from time import sleep

from lusid_workflows import ApiException


def lusidretry(fn):

    @functools.wraps(fn)
    def __retry(*args, **kwargs):

        retries = kwargs.get("lusid_retries", 3)
        if not isinstance(retries, int):
            retries = 3

        # remove the lusid_retries param from the kwargs
        kwargs.pop("lusid_retries", None)

        tries = 0
        while tries < retries+1:
            try:
                return fn(*args, **kwargs)
            except ApiException as ex:

                tries += 1
                retry_after = ex.headers.get("Retry-After")

                # have done max number of retries
                if tries == retries:
                    raise

                # try after delay
                elif retry_after is not None:
                    sleep(retry_after)

                # no retry header
                else:
                    raise

    return __retry



# coding: utf-8

"""
    FINBOURNE Workflows API

    FINBOURNE Technology  # noqa: E501

    The version of the OpenAPI document: 0.1.21
    Contact: info@finbourne.com
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from lusid_workflows.configuration import Configuration


class WorkflowInstance(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
      required_map (dict): The key is attribute name
                           and the value is whether it is 'required' or 'optional'.
    """
    openapi_types = {
        'workflow_code': 'str',
        'workflow_version': 'int',
        'workflow_instance_id': 'str',
        'correlation_id': 'str',
        'state': 'str',
        'execution_status': 'str',
        'next_activity': 'str',
        'allowed_actions': 'list[str]',
        'started_at': 'datetime',
        'finished_at': 'datetime',
        'last_step_executed_at': 'datetime',
        'execution_context': 'dict(str, object)',
        'activities': 'list[WorkflowActivitySummary]',
        'execution_info': 'str'
    }

    attribute_map = {
        'workflow_code': 'workflowCode',
        'workflow_version': 'workflowVersion',
        'workflow_instance_id': 'workflowInstanceId',
        'correlation_id': 'correlationId',
        'state': 'state',
        'execution_status': 'executionStatus',
        'next_activity': 'nextActivity',
        'allowed_actions': 'allowedActions',
        'started_at': 'startedAt',
        'finished_at': 'finishedAt',
        'last_step_executed_at': 'lastStepExecutedAt',
        'execution_context': 'executionContext',
        'activities': 'activities',
        'execution_info': 'executionInfo'
    }

    required_map = {
        'workflow_code': 'optional',
        'workflow_version': 'optional',
        'workflow_instance_id': 'optional',
        'correlation_id': 'optional',
        'state': 'optional',
        'execution_status': 'optional',
        'next_activity': 'optional',
        'allowed_actions': 'optional',
        'started_at': 'optional',
        'finished_at': 'optional',
        'last_step_executed_at': 'optional',
        'execution_context': 'optional',
        'activities': 'optional',
        'execution_info': 'optional'
    }

    def __init__(self, workflow_code=None, workflow_version=None, workflow_instance_id=None, correlation_id=None, state=None, execution_status=None, next_activity=None, allowed_actions=None, started_at=None, finished_at=None, last_step_executed_at=None, execution_context=None, activities=None, execution_info=None, local_vars_configuration=None):  # noqa: E501
        """WorkflowInstance - a model defined in OpenAPI"
        
        :param workflow_code:  Workflow Code
        :type workflow_code: str
        :param workflow_version:  Workflow definition version. When this version changes any new instance of the workflow will have the updated version number whereas  version number on existing instances will remain the same.
        :type workflow_version: int
        :param workflow_instance_id:  Workflow Instance Id which is system generated
        :type workflow_instance_id: str
        :param correlation_id:  Id of any external entity associated with the workflow instance, otherwise system generated id
        :type correlation_id: str
        :param state:  Any info about the workflow state, i.e. transition between activities
        :type state: str
        :param execution_status:  Workflow execution status
        :type execution_status: str
        :param next_activity:  Name of the next step/activity in the workflow. Empty if the workflow has already finished execution.
        :type next_activity: str
        :param allowed_actions:  Allowed actions if next activity in the workflow requires manual action, otherwise empty.
        :type allowed_actions: list[str]
        :param started_at:  Execution started at timestamp
        :type started_at: datetime
        :param finished_at:  Execution finished at timestamp. Null if the workflow isn't finished.
        :type finished_at: datetime
        :param last_step_executed_at:  Timestamp at which last activity/step in the workflow was executed. Null if the workflow is Idle.
        :type last_step_executed_at: datetime
        :param execution_context:  Data associated with a particular instance of the workflow
        :type execution_context: dict(str, object)
        :param activities:  Summary of all activities in the workflow instance. This would suggest which activities in the workflow has  been executed or are pending.
        :type activities: list[lusid_workflows.WorkflowActivitySummary]
        :param execution_info:  Any info associated with the workflow execution. This would be an error message if ExecutionStatus is Faulted for example.
        :type execution_info: str

        """  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._workflow_code = None
        self._workflow_version = None
        self._workflow_instance_id = None
        self._correlation_id = None
        self._state = None
        self._execution_status = None
        self._next_activity = None
        self._allowed_actions = None
        self._started_at = None
        self._finished_at = None
        self._last_step_executed_at = None
        self._execution_context = None
        self._activities = None
        self._execution_info = None
        self.discriminator = None

        self.workflow_code = workflow_code
        if workflow_version is not None:
            self.workflow_version = workflow_version
        self.workflow_instance_id = workflow_instance_id
        self.correlation_id = correlation_id
        self.state = state
        self.execution_status = execution_status
        self.next_activity = next_activity
        self.allowed_actions = allowed_actions
        if started_at is not None:
            self.started_at = started_at
        self.finished_at = finished_at
        self.last_step_executed_at = last_step_executed_at
        self.execution_context = execution_context
        self.activities = activities
        self.execution_info = execution_info

    @property
    def workflow_code(self):
        """Gets the workflow_code of this WorkflowInstance.  # noqa: E501

        Workflow Code  # noqa: E501

        :return: The workflow_code of this WorkflowInstance.  # noqa: E501
        :rtype: str
        """
        return self._workflow_code

    @workflow_code.setter
    def workflow_code(self, workflow_code):
        """Sets the workflow_code of this WorkflowInstance.

        Workflow Code  # noqa: E501

        :param workflow_code: The workflow_code of this WorkflowInstance.  # noqa: E501
        :type workflow_code: str
        """

        self._workflow_code = workflow_code

    @property
    def workflow_version(self):
        """Gets the workflow_version of this WorkflowInstance.  # noqa: E501

        Workflow definition version. When this version changes any new instance of the workflow will have the updated version number whereas  version number on existing instances will remain the same.  # noqa: E501

        :return: The workflow_version of this WorkflowInstance.  # noqa: E501
        :rtype: int
        """
        return self._workflow_version

    @workflow_version.setter
    def workflow_version(self, workflow_version):
        """Sets the workflow_version of this WorkflowInstance.

        Workflow definition version. When this version changes any new instance of the workflow will have the updated version number whereas  version number on existing instances will remain the same.  # noqa: E501

        :param workflow_version: The workflow_version of this WorkflowInstance.  # noqa: E501
        :type workflow_version: int
        """

        self._workflow_version = workflow_version

    @property
    def workflow_instance_id(self):
        """Gets the workflow_instance_id of this WorkflowInstance.  # noqa: E501

        Workflow Instance Id which is system generated  # noqa: E501

        :return: The workflow_instance_id of this WorkflowInstance.  # noqa: E501
        :rtype: str
        """
        return self._workflow_instance_id

    @workflow_instance_id.setter
    def workflow_instance_id(self, workflow_instance_id):
        """Sets the workflow_instance_id of this WorkflowInstance.

        Workflow Instance Id which is system generated  # noqa: E501

        :param workflow_instance_id: The workflow_instance_id of this WorkflowInstance.  # noqa: E501
        :type workflow_instance_id: str
        """

        self._workflow_instance_id = workflow_instance_id

    @property
    def correlation_id(self):
        """Gets the correlation_id of this WorkflowInstance.  # noqa: E501

        Id of any external entity associated with the workflow instance, otherwise system generated id  # noqa: E501

        :return: The correlation_id of this WorkflowInstance.  # noqa: E501
        :rtype: str
        """
        return self._correlation_id

    @correlation_id.setter
    def correlation_id(self, correlation_id):
        """Sets the correlation_id of this WorkflowInstance.

        Id of any external entity associated with the workflow instance, otherwise system generated id  # noqa: E501

        :param correlation_id: The correlation_id of this WorkflowInstance.  # noqa: E501
        :type correlation_id: str
        """

        self._correlation_id = correlation_id

    @property
    def state(self):
        """Gets the state of this WorkflowInstance.  # noqa: E501

        Any info about the workflow state, i.e. transition between activities  # noqa: E501

        :return: The state of this WorkflowInstance.  # noqa: E501
        :rtype: str
        """
        return self._state

    @state.setter
    def state(self, state):
        """Sets the state of this WorkflowInstance.

        Any info about the workflow state, i.e. transition between activities  # noqa: E501

        :param state: The state of this WorkflowInstance.  # noqa: E501
        :type state: str
        """

        self._state = state

    @property
    def execution_status(self):
        """Gets the execution_status of this WorkflowInstance.  # noqa: E501

        Workflow execution status  # noqa: E501

        :return: The execution_status of this WorkflowInstance.  # noqa: E501
        :rtype: str
        """
        return self._execution_status

    @execution_status.setter
    def execution_status(self, execution_status):
        """Sets the execution_status of this WorkflowInstance.

        Workflow execution status  # noqa: E501

        :param execution_status: The execution_status of this WorkflowInstance.  # noqa: E501
        :type execution_status: str
        """

        self._execution_status = execution_status

    @property
    def next_activity(self):
        """Gets the next_activity of this WorkflowInstance.  # noqa: E501

        Name of the next step/activity in the workflow. Empty if the workflow has already finished execution.  # noqa: E501

        :return: The next_activity of this WorkflowInstance.  # noqa: E501
        :rtype: str
        """
        return self._next_activity

    @next_activity.setter
    def next_activity(self, next_activity):
        """Sets the next_activity of this WorkflowInstance.

        Name of the next step/activity in the workflow. Empty if the workflow has already finished execution.  # noqa: E501

        :param next_activity: The next_activity of this WorkflowInstance.  # noqa: E501
        :type next_activity: str
        """

        self._next_activity = next_activity

    @property
    def allowed_actions(self):
        """Gets the allowed_actions of this WorkflowInstance.  # noqa: E501

        Allowed actions if next activity in the workflow requires manual action, otherwise empty.  # noqa: E501

        :return: The allowed_actions of this WorkflowInstance.  # noqa: E501
        :rtype: list[str]
        """
        return self._allowed_actions

    @allowed_actions.setter
    def allowed_actions(self, allowed_actions):
        """Sets the allowed_actions of this WorkflowInstance.

        Allowed actions if next activity in the workflow requires manual action, otherwise empty.  # noqa: E501

        :param allowed_actions: The allowed_actions of this WorkflowInstance.  # noqa: E501
        :type allowed_actions: list[str]
        """

        self._allowed_actions = allowed_actions

    @property
    def started_at(self):
        """Gets the started_at of this WorkflowInstance.  # noqa: E501

        Execution started at timestamp  # noqa: E501

        :return: The started_at of this WorkflowInstance.  # noqa: E501
        :rtype: datetime
        """
        return self._started_at

    @started_at.setter
    def started_at(self, started_at):
        """Sets the started_at of this WorkflowInstance.

        Execution started at timestamp  # noqa: E501

        :param started_at: The started_at of this WorkflowInstance.  # noqa: E501
        :type started_at: datetime
        """

        self._started_at = started_at

    @property
    def finished_at(self):
        """Gets the finished_at of this WorkflowInstance.  # noqa: E501

        Execution finished at timestamp. Null if the workflow isn't finished.  # noqa: E501

        :return: The finished_at of this WorkflowInstance.  # noqa: E501
        :rtype: datetime
        """
        return self._finished_at

    @finished_at.setter
    def finished_at(self, finished_at):
        """Sets the finished_at of this WorkflowInstance.

        Execution finished at timestamp. Null if the workflow isn't finished.  # noqa: E501

        :param finished_at: The finished_at of this WorkflowInstance.  # noqa: E501
        :type finished_at: datetime
        """

        self._finished_at = finished_at

    @property
    def last_step_executed_at(self):
        """Gets the last_step_executed_at of this WorkflowInstance.  # noqa: E501

        Timestamp at which last activity/step in the workflow was executed. Null if the workflow is Idle.  # noqa: E501

        :return: The last_step_executed_at of this WorkflowInstance.  # noqa: E501
        :rtype: datetime
        """
        return self._last_step_executed_at

    @last_step_executed_at.setter
    def last_step_executed_at(self, last_step_executed_at):
        """Sets the last_step_executed_at of this WorkflowInstance.

        Timestamp at which last activity/step in the workflow was executed. Null if the workflow is Idle.  # noqa: E501

        :param last_step_executed_at: The last_step_executed_at of this WorkflowInstance.  # noqa: E501
        :type last_step_executed_at: datetime
        """

        self._last_step_executed_at = last_step_executed_at

    @property
    def execution_context(self):
        """Gets the execution_context of this WorkflowInstance.  # noqa: E501

        Data associated with a particular instance of the workflow  # noqa: E501

        :return: The execution_context of this WorkflowInstance.  # noqa: E501
        :rtype: dict(str, object)
        """
        return self._execution_context

    @execution_context.setter
    def execution_context(self, execution_context):
        """Sets the execution_context of this WorkflowInstance.

        Data associated with a particular instance of the workflow  # noqa: E501

        :param execution_context: The execution_context of this WorkflowInstance.  # noqa: E501
        :type execution_context: dict(str, object)
        """

        self._execution_context = execution_context

    @property
    def activities(self):
        """Gets the activities of this WorkflowInstance.  # noqa: E501

        Summary of all activities in the workflow instance. This would suggest which activities in the workflow has  been executed or are pending.  # noqa: E501

        :return: The activities of this WorkflowInstance.  # noqa: E501
        :rtype: list[lusid_workflows.WorkflowActivitySummary]
        """
        return self._activities

    @activities.setter
    def activities(self, activities):
        """Sets the activities of this WorkflowInstance.

        Summary of all activities in the workflow instance. This would suggest which activities in the workflow has  been executed or are pending.  # noqa: E501

        :param activities: The activities of this WorkflowInstance.  # noqa: E501
        :type activities: list[lusid_workflows.WorkflowActivitySummary]
        """

        self._activities = activities

    @property
    def execution_info(self):
        """Gets the execution_info of this WorkflowInstance.  # noqa: E501

        Any info associated with the workflow execution. This would be an error message if ExecutionStatus is Faulted for example.  # noqa: E501

        :return: The execution_info of this WorkflowInstance.  # noqa: E501
        :rtype: str
        """
        return self._execution_info

    @execution_info.setter
    def execution_info(self, execution_info):
        """Sets the execution_info of this WorkflowInstance.

        Any info associated with the workflow execution. This would be an error message if ExecutionStatus is Faulted for example.  # noqa: E501

        :param execution_info: The execution_info of this WorkflowInstance.  # noqa: E501
        :type execution_info: str
        """

        self._execution_info = execution_info

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, WorkflowInstance):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, WorkflowInstance):
            return True

        return self.to_dict() != other.to_dict()



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright © 2018 FINBOURNE TECHNOLOGY LTD
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------------


from setuptools import setup, find_packages  # noqa: H301

# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

REQUIRES = [
    "urllib3 >= 1.15",
    "six >= 1.10",
    "certifi >= 14.05.14",
    "python-dateutil >= 2.5.3",
    "urllib3 >= 1.15.1",
    "requests >= 2.21.0"
]

version = {}
with open("./lusid_workflows/__version__.py") as fp:
    exec(fp.read(), version)

setup(

    name='lusid-workflows-sdk-preview',
    version=version['__version__'],
    description='Python Preview SDK for Workflows API',
    url='https://github.com/finbourne/workflows-sdk-python-preview',
    author='FINBOURNE Technology',
    author_email='engineering@finbourne.com',
    license='MIT',
    keywords=["OpenAPI", "FINBOURNE", "LUSID", "Notifications API"],
    install_requires=REQUIRES,
    packages=find_packages(exclude=['tests*']),
    include_package_data=True
)



class ApiConfiguration:

    def __init__(self, token_url=None, api_url=None, username=None, password=None, client_id=None, client_secret=None,
                 app_name=None, certificate_filename=None, proxy_config=None):
        """
        The configuration required to access LUSID, read more at https://support.finbourne.com/getting-started-with-apis-sdks

        :param str token_url: The token URL of the identity provider
        :param str api_url: The API URL for the LUSID client
        :param str username: The username to use
        :param str password: The password to use
        :param str client_id: The client id to use
        :param str client_secret: The client secret to use
        :param str app_name: The name of the application calling LUSID
        :param str certificate_filename: Name of the certificate file (.pem, .cer or .crt)
        :param lusid.utilities.ProxyConfig proxy_config: The proxy configuration to use
        """
        self.__token_url = token_url
        self.__api_url = api_url
        self.__username = username
        self.__password = password
        self.__client_id = client_id
        self.__client_secret = client_secret
        self.__app_name = app_name
        self.__certificate_filename = certificate_filename
        self.__proxy_config = proxy_config

    @property
    def token_url(self):
        return self.__token_url

    @token_url.setter
    def token_url(self, value):
        self.__token_url = value

    @property
    def api_url(self):
        return self.__api_url

    @api_url.setter
    def api_url(self, value):
        self.__api_url = value

    @property
    def username(self):
        return self.__username

    @username.setter
    def username(self, value):
        self.__username = value

    @property
    def password(self):
        return self.__password

    @password.setter
    def password(self, value):
        self.__password = value

    @property
    def client_id(self):
        return self.__client_id

    @client_id.setter
    def client_id(self, value):
        self.__client_id = value

    @property
    def client_secret(self):
        return self.__client_secret

    @client_secret.setter
    def client_secret(self, value):
        self.__client_secret = value

    @property
    def app_name(self):
        return self.__app_name

    @app_name.setter
    def app_name(self, value):
        self.__app_name = value

    @property
    def certificate_filename(self):
        return self.__certificate_filename

    @certificate_filename.setter
    def certificate_filename(self, value):
        self.__certificate_filename = value

    @property
    def proxy_config(self):
        return self.__proxy_config

    @proxy_config.setter
    def proxy_config(self, value):
        self.__proxy_config = value



from urllib3 import make_headers
import os

from lusid_workflows import Configuration, ApiClient

from .api_configuration_loader import ApiConfigurationLoader
from .refreshing_token import RefreshingToken


class ApiClientBuilder:
    """
    The ApiClientBuilder is responsible for building a lusid.ApiClient. This includes obtaining an access token from
    Okta or using the provided token.

    Any validation on the inputs required to build a lusid.ApiClient is the responsibility of this ApiClientBuilder.
    """

    @staticmethod
    def __check_required_fields(object_to_check, fields):
        """
        This function checks that the provided fields on an object are populated with values other than None

        :param object_to_check: The object to check the fields (a.k.a attributes) of
        :param list[str] fields: The fields to check on the object

        :return: None
        """
        # Check for fields which have a value of None
        missing_fields = [field for field in fields if getattr(object_to_check, field) is None]

        # Raise an error if any fields have a value of None
        if len(missing_fields) > 0:
            raise ValueError(
                f"The fields {str(missing_fields)} on the {object_to_check.__class__.__name__} are set to None, "
                f"please ensure that you have provided them directly, via a secrets file or environment "
                f"variables")

    @classmethod
    def build(cls, api_secrets_filename=None, id_provider_response_handler=None, api_configuration=None,
              token=None, correlation_id=None, tcp_keep_alive=False):
        """
        :param str api_secrets_filename: The full path to the JSON file containing the API credentials and optional proxy details
        :param typing.callable id_provider_response_handler: An optional function to handle the Okta response
        :param lusid.utilities.ApiConfiguration api_configuration: A pre-populated ApiConfiguration
        :param str token: The pre-populated access token to use instead of asking Okta for a token
        :param str correlation_id: Correlation id for all calls made from the returned ApiClient instance, added as a header to each request
        :param bool tcp_keep_alive: A flag that controls if the API client uses tcp keep-alive probes

        :return: lusid.ApiClient: The configured LUSID ApiClient
        """

        # Load the configuration
        configuration = ApiConfigurationLoader.load(api_secrets_filename)

        # If an api_configuration has been provided override the loaded configuration with any properties that it has
        if api_configuration is not None:
            for key, value in vars(api_configuration).items():
                if value is not None:
                    setattr(configuration, key, value)

        # Use the access token provided if it exists
        if token is not None:
            # Check that there is an api_url available
            cls.__check_required_fields(configuration, ["api_url"])
            api_token = token
        # Otherwise generate an access token from Okta and use a RefreshingToken going forward
        else:
            # Check that all the required fields for generating a token exist
            cls.__check_required_fields(configuration, [
                "api_url",
                "password",
                "username",
                "client_id",
                "client_secret",
                "token_url"])

            # Generate an access token
            api_token = RefreshingToken(
                api_configuration=configuration,
                id_provider_response_handler=id_provider_response_handler
            )

        # Initialise the API client using the token so that it can be included in all future requests
        config = Configuration(tcp_keep_alive=tcp_keep_alive)
        config.access_token = api_token
        config.host = configuration.api_url

        # Set the certificate from the configuration
        config.ssl_ca_cert = configuration.certificate_filename

        # Set the proxy for LUSID if needed
        if configuration.proxy_config is not None:
            config.proxy = configuration.proxy_config.address
            if configuration.proxy_config.username is not None and configuration.proxy_config.password is not None:
                config.proxy_headers = make_headers(
                    proxy_basic_auth=f"{configuration.proxy_config.username}:{configuration.proxy_config.password}"
                )

        # Create and return the ApiClient
        api_client = ApiClient(configuration=config)

        # set the application name if specified
        if configuration.app_name is not None:
            api_client.set_default_header("X-LUSID-Application", configuration.app_name)

        # set a correlation id for all requests initiated with this ApiClient
        corr_id = correlation_id or os.getenv("FBN_CORRELATION_ID")
        if corr_id is not None:
            api_client.set_default_header("CorrelationId", corr_id)

        return api_client
